# -*- coding: utf-8 -*-
"""mls.apiclient exception classes."""


class MLSError(Exception):
    """Main exception class."""
    pass


class ImproperlyConfigured(MLSError):
    """This exception is raised on configuration errors."""
    pass


class ObjectNotFound(MLSError):
    """This exception is raised if an object can not be found."""
    pass


class MultipleResults(MLSError):
    """This exception is raised on multiple results.

    That is, if a get request returns more than one result.
    """
    def __str__(self):
        return 'Your query had multiple results.'


class MissingParam(TypeError):
    pass


class ConnectionError(Exception):
    """Base exception for any type of connection error with the requests."""

    def __init__(self, status_code, reason=None, url=None):
        self.status_code = status_code
        self.reason = reason
        self.url = url

    def __str__(self):
        message = 'Response status: {0}.'.format(self.status_code)
        if self.reason:
            message += ' Reason: {0}.'.format(self.reason)
        if self.url:
            message += ' URL: {0}'.format(self.url)
        return message


class Redirection(ConnectionError):
    """3xx Redirection."""
    pass


class ClientError(ConnectionError):
    """4xx Client Error."""
    pass


class BadRequest(ClientError):
    """400 Bad Request."""
    pass


class UnauthorizedAccess(ClientError):
    """401 Unauthorized."""
    pass


class ResourceNotFound(ClientError):
    """404 Not Found."""
    pass


class ServerError(ConnectionError):
    """5xx Server Error."""
    pass



# -*- coding: utf-8 -*-
"""Test the resource classes."""

# python imports
import httpretty
import json

# local imports
from mls.apiclient import api, resources
from mls.apiclient.tests import base, utils
from mls.apiclient.utils import join_url


class ResourceTestCase(base.BaseTestCase):
    """Test resource class."""

    def setUp(self):
        self.api = api.API(self.BASE_URL)
        self.data = json.loads(utils.load_fixture('basic_single.json'))

    def _callFUT(self, api, data):
        return resources.Resource(api, data)

    def test_wrong_data(self):
        """Validate the ValueError if data is not a dictionary."""
        self.assertRaises(ValueError, self._callFUT, self.api, None)
        self.assertRaises(ValueError, self._callFUT, self.api, 1)
        self.assertRaises(ValueError, self._callFUT, self.api, 1.1)
        self.assertRaises(ValueError, self._callFUT, self.api, True)
        self.assertRaises(ValueError, self._callFUT, self.api, 'Foo')
        self.assertRaises(ValueError, self._callFUT, self.api, u'Foo')

    def test_correct_data(self):
        """Validate the correct initialization with a dictionary."""
        data = {
            'response': {
                'foo': 'bar',
            },
        }
        foo = self._callFUT(self.api, data)
        self.assertIsInstance(foo, resources.Resource)
        self.assertEqual(data.get('response'), foo._data)

    def test_attributes(self):
        """Validate the content of the data dictionary."""
        data = {
            'response': {
                'foo': 'bar',
            },
        }
        foo = self._callFUT(self.api, data)
        self.assertEqual(foo.foo, 'bar')

    def test_missing_attributes(self):
        """Validate that an error is raised when an attribute is missing."""
        data = {
            'response': {
                'foo': 'bar',
            },
        }
        foo = self._callFUT(self.api, data)
        self.assertRaises(AttributeError, getattr, foo, 'bar')

    def test_get_attributes(self):
        """Validate the output of the get_attributes method."""
        data = {
            'response': {
                'foo': 'bar',
                'baz': 'barbar',
            },
        }
        foo = self._callFUT(self.api, data)
        result = foo.get_attributes()
        self.assertIn('foo', result)
        self.assertIn('baz', result)
        self.assertEqual(len(result), 2)

        foo = self._callFUT(self.api, self.data)
        result = foo.get_attributes()
        self.assertIn('id', result)
        self.assertIn('links', result)
        self.assertEqual(len(result), 2)

    def test_get_id(self):
        """Validate the output of the get_id method."""
        foo = self._callFUT(self.api, {})
        self.assertIsNone(foo.get_id())

        data = {
            'response': {
                'id': 'my_id',
            },
        }
        foo = self._callFUT(self.api, data)
        self.assertEqual(foo.get_id(), 'my_id')

        foo = self._callFUT(self.api, self.data)
        self.assertEqual(foo.get_id(), 'test_id')

    def test_get_url(self):
        """Validate the output of the get_url method."""
        foo = self._callFUT(self.api, {})
        self.assertIsNone(foo.get_url())

        data = {
            'response': {
                'links': [{
                    'href': 'http://example.com',
                    'rel': 'next',
                    'method': 'GET',
                }]
            }
        }
        foo = self._callFUT(self.api, data)
        self.assertIsNone(foo.get_url())

        data = {
            'response': {
                'links': [{
                    'href': 'http://example.com',
                    'rel': 'self',
                    'method': 'GET',
                }]
            }
        }
        foo = self._callFUT(self.api, data)
        self.assertEqual(foo.get_url(), 'http://example.com')

        foo = self._callFUT(self.api, self.data)
        self.assertEqual(
            foo.get_url(),
            'http://demomls.com/api/rest/v1/test_url',
        )

    def test_headers(self):
        """Validate the headers of the response."""
        foo = self._callFUT(self.api, {})
        self.assertEqual(foo.get_headers(), {})

        data = {
            'headers': {
                'header-1': 'value1',
                'header-2': 'value2',
            },
        }
        foo = self._callFUT(self.api, data)
        headers = foo.get_headers()
        self.assertIn('header-1', headers)
        self.assertIn('header-2', headers)
        self.assertEqual(len(headers), 2)

        foo = self._callFUT(self.api, self.data)
        headers = foo.get_headers()
        self.assertIn('result-single', headers)
        self.assertTrue(headers.get('result-single'))
        self.assertEqual(len(headers), 1)


class AgencyTestCase(base.BaseTestCase):
    """Test 'Agency' resource class."""

    def setUp(self):
        self.api = api.API(self.BASE_URL)

    def _callFUT(self, api, data):
        return resources.Agency(api, data)

    def test_listings(self):
        """Validate the listing search for agencies."""
        agency = self._callFUT(self.api, {})
        self.assertRaises(NotImplementedError, agency.listings)

    def test_developments(self):
        """Validate the development search for agencies."""
        agency = self._callFUT(self.api, {})
        self.assertRaises(NotImplementedError, agency.developments)


class AgentTestCase(base.BaseTestCase):
    """Test 'Agent' resource class."""

    def setUp(self):
        self.api = api.API(self.BASE_URL)

    def _callFUT(self, api, data):
        return resources.Agent(api, data)

    def test_listings(self):
        """Validate the listing search for agents."""
        agent = self._callFUT(self.api, {})
        self.assertRaises(NotImplementedError, agent.listings)


class DevelopmentTestCase(base.BaseTestCase):
    """Test 'Development' resource class."""

    endpoint = 'developments'

    def setUp(self):
        self.api = api.API(self.BASE_URL)

    def _callFUT(self, api, data):
        return resources.Development(api, data)

    @httpretty.httprettified
    def test_get_development(self):
        """Validate the 'get' endpoint."""
        dev_id = 'dev-agency__dev001'
        resource = join_url(self.endpoint, dev_id)
        response = utils.load_fixture('development_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.Development.get(self.api, dev_id)
        self.assertIsInstance(result, resources.Development)
        response_dict = json.loads(response)
        self.assertEqual(result._data, response_dict.get('response'))

    @httpretty.httprettified
    def test_get_all(self):
        """Validate the 'search' endpoint to get all developments."""
        resource = self.endpoint
        response = utils.load_fixture('development_list_1.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.Development.search(self.api)
        self.assertIsInstance(result, resources.Development)
        response_dict = json.loads(response)
        self.assertEqual(result._data, response_dict.get('response'))

    def test_get_items_development(self):
        """Validate the 'search' endpoint to get all developments and get the
        list of developments from the result.
        """
        data_single = json.loads(utils.load_fixture('development_en.json'))
        development = self._callFUT(self.api, data_single)
        items = development.get_items()
        self.assertEqual(len(items), 0)

        data_list = json.loads(utils.load_fixture('development_list_1.json'))
        development = self._callFUT(self.api, data_list)
        items = development.get_items()
        self.assertEqual(len(items), 1)
        for item in items:
            self.assertIsInstance(item, resources.Development)
            self.assertEqual(item._data, data_single.get('response'))

    @httpretty.httprettified
    def test_development_fields(self):
        """Validate the 'field_titles' endpoint."""
        resource = join_url('field_titles', self.endpoint)
        response = utils.load_fixture('development_fields_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.Development.get_field_titles(self.api)
        self.assertEqual(result, json.loads(response))

    @httpretty.httprettified
    def test_development_field_order(self):
        """Validate the 'field_order' endpoint."""
        resource = join_url('field_order', self.endpoint)
        response = utils.load_fixture('development_fields_order.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.Development.get_field_order(self.api)
        self.assertEqual(result, json.loads(response))

    def test_listings(self):
        """Validate the listing search for developments."""
        development = self._callFUT(self.api, {})
        self.assertIsNone(development.listings())

    def test_listings_url(self):
        """Validate the deprecated listing search for developments."""
        data = json.loads(utils.load_fixture('development_en.json'))
        development = self._callFUT(self.api, data)
        self.assertEqual(
            development.listings_url,
            'http://demomls.com/api/listings/search?lang=en'
            '&apikey=YOUR_API_KEY&development_listings=dev-agency__dev001',
        )

    def test_pictures(self):
        """Validate the pictures for developments."""
        data = json.loads(utils.load_fixture('development_en.json'))
        development = self._callFUT(self.api, data)
        images = development.pictures()
        self.assertIsInstance(images, list)
        self.assertEqual(len(images), 2)
        for image in images:
            self.assertIsInstance(image, resources.Image)

    @httpretty.httprettified
    def test_property_groups(self):
        """Validate the property group search for developments."""
        data = json.loads(utils.load_fixture('development_en.json'))
        development = self._callFUT(self.api, data)

        resource = join_url(self.endpoint, development.get_id(), 'groups')
        response = utils.load_fixture('group_list_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        groups = development.groups()
        self.assertEqual(len(groups), 2)
        for group in groups:
            self.assertIsInstance(group, resources.PropertyGroup)

    @httpretty.httprettified
    def test_development_phases(self):
        """Validate the development phase search for developments."""
        data = json.loads(utils.load_fixture('development_en.json'))
        development = self._callFUT(self.api, data)

        resource = join_url(self.endpoint, development.get_id(), 'phases')
        response = utils.load_fixture('phase_list_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        phases = development.phases()
        self.assertEqual(len(phases), 2)
        for phase in phases:
            self.assertIsInstance(phase, resources.DevelopmentPhase)


class DevelopmentPhaseTestCase(base.BaseTestCase):
    """Test 'Development Phase' resource class."""

    endpoint = 'development_phases'

    def setUp(self):
        self.api = api.API(self.BASE_URL)

    def _callFUT(self, api, data):
        return resources.DevelopmentPhase(api, data)

    @httpretty.httprettified
    def test_phase_fields(self):
        """Validate the 'field_titles' endpoint."""
        resource = join_url('field_titles', self.endpoint)
        response = utils.load_fixture('phase_fields_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.DevelopmentPhase.get_field_titles(self.api)
        self.assertEqual(result, json.loads(response))

    @httpretty.httprettified
    def test_phase_field_order(self):
        """Validate the 'field_order' endpoint."""
        resource = join_url('field_order', self.endpoint)
        response = utils.load_fixture('phase_fields_order.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.DevelopmentPhase.get_field_order(self.api)
        self.assertEqual(result, json.loads(response))

    @httpretty.httprettified
    def test_get_all(self):
        """Validate the 'search' endpoint to get all phases."""
        resource = self.endpoint
        response = utils.load_fixture('phase_list_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.DevelopmentPhase.search(self.api)
        self.assertIsInstance(result, resources.DevelopmentPhase)
        response_dict = json.loads(response)
        self.assertEqual(result._data, response_dict.get('response'))

    @httpretty.httprettified
    def test_get_items_phases(self):
        """Validate the 'search' endpoint to get all phases and get the
        list of phases from the result.
        """
        resource = self.endpoint
        response = utils.load_fixture('phase_list_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.DevelopmentPhase.search(self.api)
        items = result.get_items()
        self.assertEqual(len(items), 2)
        for item in items:
            self.assertIsInstance(item, resources.DevelopmentPhase)
            self.assertIn('title', item._data)
            self.assertIn('units', item._data)
            self.assertIn('unit_types', item._data)
            self.assertIn('status', item._data)
            self.assertIn('listings_url', item._data)

    def test_listings(self):
        """Validate the listing search for development phases."""
        phase = self._callFUT(self.api, {})
        self.assertIsNone(phase.listings())

    def test_listings_url(self):
        """Validate the deprecated listing search for development phases."""
        data = json.loads(utils.load_fixture('phase_list_en.json'))
        phase_list = self._callFUT(self.api, data)
        phase = phase_list.get_items()[0]
        self.assertEqual(
            phase.listings_url,
            'http://demomls.com/api/listings/search?'
            'apikey=YOUR_API_KEY&phase_listings=dev-agency__dev001__phase-1',
        )


class ListingTestCase(base.BaseTestCase):
    """Test 'Listing' resource class."""

    def setUp(self):
        self.api = api.API(self.BASE_URL)

    def _callFUT(self, api, data):
        return resources.Listing(api, data)

    def test_pictures(self):
        """Validate the pictures for listings."""
        listing = self._callFUT(self.api, {})
        self.assertRaises(NotImplementedError, listing.pictures)


class PropertyGroupTestCase(base.BaseTestCase):
    """Test 'Property Group' resource class."""

    endpoint = 'property_groups'

    def setUp(self):
        self.api = api.API(self.BASE_URL)

    def _callFUT(self, api, data):
        return resources.PropertyGroup(api, data)

    @httpretty.httprettified
    def test_group_fields(self):
        """Validate the 'field_titles' endpoint."""
        resource = join_url('field_titles', self.endpoint)
        response = utils.load_fixture('group_fields_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.PropertyGroup.get_field_titles(self.api)
        self.assertEqual(result, json.loads(response))

    @httpretty.httprettified
    def test_group_field_order(self):
        """Validate the 'field_order' endpoint."""
        resource = join_url('field_order', self.endpoint)
        response = utils.load_fixture('group_fields_order.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.PropertyGroup.get_field_order(self.api)
        self.assertEqual(result, json.loads(response))

    @httpretty.httprettified
    def test_get_all(self):
        """Validate the 'search' endpoint to get all property groups."""
        resource = self.endpoint
        response = utils.load_fixture('group_list_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.PropertyGroup.search(self.api)
        self.assertIsInstance(result, resources.PropertyGroup)
        response_dict = json.loads(response)
        self.assertEqual(result._data, response_dict.get('response'))

    @httpretty.httprettified
    def test_get_items_groups(self):
        """Validate the 'search' endpoint to get all groups and get the
        list of groups from the result.
        """
        resource = self.endpoint
        response = utils.load_fixture('group_list_en.json')
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.API_BASE, resource),
            body=response,
        )
        result = resources.PropertyGroup.search(self.api)
        items = result.get_items()
        self.assertEqual(len(items), 2)
        for item in items:
            self.assertIsInstance(item, resources.PropertyGroup)
            self.assertIn('title', item._data)
            self.assertIn('status', item._data)
            self.assertIn('listings_url', item._data)

    def test_listings(self):
        """Validate the listing search for property groups."""
        group = self._callFUT(self.api, {})
        self.assertIsNone(group.listings())

    def test_listings_url(self):
        """Validate the deprecated listing search for property groups."""
        data = json.loads(utils.load_fixture('group_list_en.json'))
        group_list = self._callFUT(self.api, data)
        group = group_list.get_items()[0]
        self.assertEqual(
            group.listings_url,
            'http://demomls.com/api/listings/search?property_group_listings'
            '=dev-agency__dev001__classy&apikey=YOUR_API_KEY'
        )


class ImageTestCase(base.BaseTestCase):
    """Test 'Image' resource class."""

    def setUp(self):
        self.data = {
            'large': 'http://demo.com/pics/img1/@@scaling/img1-large.jpeg',
            'tile': 'http://demo.com/pics/img1/@@scaling/img1-tile.jpeg',
            'thumb': 'http://demo.com/pics/img1/@@scaling/img1-thumb.jpeg',
            'title': 'Title of image 1',
            'url': 'http://demo.com/pics/img1/@@scaling/img1-url.jpeg',
            'preview': 'http://demo.com/pics/img1/@@scaling/img1-preview.jpeg',
            'id': 'img1',
            'description': 'Description of image 1',
        }

    def _callFUT(self, data):
        return resources.Image(data)

    def test_title(self):
        """Validate the 'title' attribute of Images."""
        image = self._callFUT(self.data)
        self.assertEqual(image.title, 'Title of image 1')

    def test_id(self):
        """Validate the 'id' attribute of Images."""
        image = self._callFUT(self.data)
        self.assertEqual(image.id, 'img1')

    def test_description(self):
        """Validate the 'description' attribute of Images."""
        image = self._callFUT(self.data)
        self.assertEqual(image.description, 'Description of image 1')

    def test_get_large(self):
        """Validate getting the 'large' scale url."""
        image = self._callFUT(self.data)
        self.assertEqual(
            image.get('large'),
            'http://demo.com/pics/img1/@@scaling/img1-large.jpeg',
        )

    def test_get_tile(self):
        """Validate getting the 'tile' scale url."""
        image = self._callFUT(self.data)
        self.assertEqual(
            image.get('tile'),
            'http://demo.com/pics/img1/@@scaling/img1-tile.jpeg',
        )

    def test_get_thumb(self):
        """Validate getting the 'thumb' scale url."""
        image = self._callFUT(self.data)
        self.assertEqual(
            image.get('thumb'),
            'http://demo.com/pics/img1/@@scaling/img1-thumb.jpeg',
        )

    def test_get_url(self):
        """Validate getting the 'url' scale url."""
        image = self._callFUT(self.data)
        self.assertEqual(
            image.get('url'),
            'http://demo.com/pics/img1/@@scaling/img1-url.jpeg',
        )

    def test_get_preview(self):
        """Validate getting the 'preview' scale url."""
        image = self._callFUT(self.data)
        self.assertEqual(
            image.get('preview'),
            'http://demo.com/pics/img1/@@scaling/img1-preview.jpeg',
        )

    def test_get_invalid(self):
        """Validate getting a non-existent scale url."""
        image = self._callFUT(self.data)
        self.assertIsNone(image.get('invalid'))

    def test_get_scale_title(self):
        """Validate the behavior of trying to get the 'title' scale url."""
        image = self._callFUT(self.data)
        self.assertIsNone(image.get('title'))

    def test_get_scale_id(self):
        """Validate the behavior of trying to get the 'id' scale url."""
        image = self._callFUT(self.data)
        self.assertIsNone(image.get('id'))

    def test_get_scale_description(self):
        """Validate the behavior of trying to get the 'description' scale url.
        """
        image = self._callFUT(self.data)
        self.assertIsNone(image.get('description'))



# -*- coding: utf-8 -*-
"""Test all the resources."""

# python imports
import httpretty

# local imports
from mls.apiclient import api, resources
from mls.apiclient.tests import base


class IntegrationTestCase(base.BaseTestCase):
    """Test resource class."""

    def setUp(self):
        self.api = api.API(
            self.BASE_URL,
            api_key='YOUR_API_KEY',
            lang='en',
            debug=True,
        )

    @httpretty.httprettified
    def test_development_list(self):
        """Development integration test."""
        self.setup_integration_test()
        developments = resources.Development.search(self.api)
        development_list = developments.get_items()
        self.assertEqual(len(development_list), 25)



# -*- coding: utf-8 -*-
"""MLS rest client entity resource classes."""

# python imports
import urlparse

# local imports
from mls.apiclient import (
    REST_API_URL,
    REST_API_VERSION,
    client,
    utils,
)

IMG_FIELDS = ('id', 'title', 'description', )


class Image(object):
    """Image resource."""

    def __init__(self, data):
        self._data = data

    @property
    def title(self):
        return self._data.get('title')

    @property
    def id(self):
        return self._data.get('id')

    @property
    def description(self):
        return self._data.get('description')

    def get(self, scale='url'):
        if scale in IMG_FIELDS:
            return
        return self._data.get(scale)


class Resource(object):
    """Base class for resources."""

    endpoint = None

    def __init__(self, api, data):
        if not isinstance(data, dict):
            raise ValueError(
                'Data must be dictionary with content of the resource.'
            )

        self._api = api
        self._headers = data.get('headers', {})
        self._data = data.get('response', {})
        self._links = self._data.get('links', {})

    def __getattr__(self, name):
        """Returns a data attribute or raises AttributeError."""
        try:
            return self._data[name]
        except KeyError:
            return object.__getattribute__(self, name)

    @classmethod
    def get(cls, api, resource_id):
        """Returns one object of this Resource.

        You have to give one keyword argument to find the object.
        """
        url = utils.join_url(cls.get_endpoint_url(), resource_id)
        return cls(api, api.get(url))

    @classmethod
    def search(cls, api, params=None):
        """Returns a list of objects with optional search parameters.

        You can search for objects by giving one or more keyword arguments.
        Use limit and offset to limit the results.
        """
        url = cls.get_endpoint_url()
        return cls(api, api.get(url, params))

    @classmethod
    def get_field_titles(cls, api):
        """Return the translated titles of the fields."""
        url = utils.join_url(
            REST_API_URL,
            REST_API_VERSION,
            'field_titles',
            cls.endpoint,
        )
        return api.get(url)

    @classmethod
    def get_field_order(cls, api):
        """Return the list of fieldnames in order as defined in the MLS."""
        url = utils.join_url(
            REST_API_URL,
            REST_API_VERSION,
            'field_order',
            cls.endpoint,
        )
        return api.get(url)

    @classmethod
    def get_endpoint_url(cls):
        """Returns the URL to the resource object."""
        return utils.join_url(REST_API_URL, REST_API_VERSION, cls.endpoint)

    def get_attributes(self):
        """Returns a list of all attributes of a Resource object."""
        return self._data.keys()

    def get_id(self):
        """Returns the id of the resource object."""
        return self._data.get('id', None)

    def get_url(self):
        """Returns the URL to the resource object."""
        return utils.get_link(self._links, 'self')

    def get_headers(self):
        """Returns a dictionary of the headers for this resource object."""
        return self._headers

    def get_items(self):
        """Returns the list of Resource objects when the data contains a
        collection.
        """
        result = []
        for item in self._data.get('collection', []):
            data = utils.wrap_data_response(item)
            result.append(self.__class__(self._api, data))
        return result

    def _return_value(self, fields, data):
        return dict([(
            f, {'label': fields.get(f, f), 'value': data.get(f, None)}
        ) for f in data.keys()])


class Agency(Resource):
    """'Agency' entity resource class."""

    endpoint = 'agencies'

    def listings(self, params=None):
        """Search for listings within that agency."""
        raise NotImplementedError

    def developments(self):
        """Search for developments within that agency."""
        raise NotImplementedError


class Agent(Resource):
    """'Agent' entity resource class."""

    endpoint = 'agents'

    def listings(self, params=None):
        """Search for listings for that agent."""
        raise NotImplementedError


class Development(Resource):
    """'Development Project' entity resource class."""

    endpoint = 'developments'

    def __init__(self, api, data):
        super(Development, self).__init__(api, data)
        self.__class_agency__ = Agency
        self.__class_agent__ = Agent
        self.__class_group__ = PropertyGroup
        self.__class_phase__ = DevelopmentPhase

    def agency(self):
        """Return the responsible agency for this development."""
        value = self._data.get('agency', None)
        if value is not None:
            data = {
                'response': value,
            }
            return self.__class_agency__(self._api, data)

    def agent(self):
        """Return the responsible agent for this development."""
        value = self._data.get('agent', None)
        if value is not None:
            data = {
                'response': value,
            }
            return self.__class_agent__(self._api, data)

    def groups(self, params=None):
        """Search for property groups within that development."""
        url = self._data.get('groups')
        data = self._api.request(url, 'GET', params=params)
        return self.__class_group__(self._api, data).get_items()

    def listings(self, params=None):
        """Search for listings assigned to that development project."""
        if params is None:
            params = {}
        url = self._data.get('listing_url', None)
        if url is None:
            return
        url_params = url.split('?')
        if len(url_params) < 2:
            return
        url_params = dict(urlparse.parse_qsl(url_params[1]))
        params.update(url_params)
        listing_resource = client.ListingResource(
            self._api.base_url, api_key=self._api.api_key,
        )
        return listing_resource.search(params=params)

    def phases(self, params=None):
        """Search for development phases within that development."""
        url = self._data.get('phases')
        data = self._api.request(url, 'GET', params=params)
        return self.__class_phase__(self._api, data).get_items()

    def pictures(self):
        """Get the pictures for that development."""
        result = []
        items = self._data.get('pictures', [])
        if items is None:
            return
        for item in items:
            result.append(Image(item))
        return result

    def representatives(self):
        """Get the representatives for that development."""
        result = []
        items = self._data.get('representatives', [])
        if items is None:
            return
        for item in items:
            data = {
                'response': item,
            }
            result.append(self.__class_agent__(self._api, data))
        return result


class DevelopmentPhase(Resource):
    """'Development Phase' entity resource class."""

    endpoint = 'development_phases'

    def listings(self, params=None):
        """Search for listings assigned to that development phase."""
        if params is None:
            params = {}
        url = self._data.get('listing_url', None)
        if url is None:
            return
        url_params = url.split('?')
        if len(url_params) < 2:
            return
        url_params = dict(urlparse.parse_qsl(url_params[1]))
        params.update(url_params)
        listing_resource = client.ListingResource(
            self._api.base_url, api_key=self._api.api_key,
        )
        return listing_resource.search(params=params)


class Listing(Resource):
    """'Listing' entity resource class."""

    def pictures(self):
        """Get the pictures for that listing."""
        raise NotImplementedError


class PropertyGroup(Resource):
    """'Property Group' entity resource class."""

    endpoint = 'property_groups'

    def listings(self, params=None):
        """Search for listings assigned to that property group."""
        if params is None:
            params = {}
        url = self._data.get('listing_url', None)
        if url is None:
            return
        url_params = url.split('?')
        if len(url_params) < 2:
            return
        url_params = dict(urlparse.parse_qsl(url_params[1]))
        params.update(url_params)
        listing_resource = client.ListingResource(
            self._api.base_url, api_key=self._api.api_key,
        )
        return listing_resource.search(params=params)



# -*- coding: utf-8 -*-
"""Test Propertyshelf MLS listings API."""

# python imports
import httpretty

# local imports
from mls.apiclient.client import ListingResource
from mls.apiclient.tests import base, utils


class TestCategories(base.BaseTestCase):
    """Listing categories test case."""
    PATH = '/api/listings/categories'

    @httpretty.httprettified
    def test_listing_types(self):
        resource = 'listing_types'
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.URL, resource),
            body=utils.load_fixture('category_listing_types_en.json')
        )
        client = ListingResource(self.BASE_URL)
        category = client.category(resource)
        assert httpretty.last_request().querystring == {
            'format': ['json'],
        }
        expected = [
            ('ll', 'Land Listing'),
            ('rl', 'Residential Lease'),
            ('rs', 'Residential Sale'),
        ]
        assert category == expected

    @httpretty.httprettified
    def test_view_types(self):
        resource = 'view_types'
        httpretty.register_uri(
            httpretty.GET,
            utils.get_url(self.URL, resource),
            body=utils.load_fixture('category_view_types_en.json')
        )
        client = ListingResource(self.BASE_URL)
        category = client.category(resource)
        assert httpretty.last_request().querystring == {
            'format': ['json'],
        }
        expected = [
            ('beach_view', 'Beach View'),
        ]
        assert category == expected



# -*- coding: utf-8 -*-
"""Several test utils."""

# python imports
import os


def get_url(base_url, endpoint):
    return '/'.join([
        base_url,
        endpoint,
    ])


def load_fixture(name):
    """Return a file-like fixture, just like urlopen would."""
    fixture = open(
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'fixtures',
            name,
        ), 'r'
    )
    return fixture.read()


def wrap_content(content, status_code=200, headers={}):
    """Wrap string JSON content in the appropriate response format with the
    given status code and return the JSON response as a string.
    """
    return u'{{"status": {0}, "headers": {1}, "response": {2}}}'.format(
        status_code,
        headers,
        content,
    )



# -*- coding: utf-8 -*-
"""Test case base class."""

# python imports
import httpretty
import itertools
import re
import urllib
try:
    import unittest2 as unittest
except ImportError:
    import unittest

# local imports
from mls.apiclient import (
    REST_API_URL,
    REST_API_VERSION,
)
from mls.apiclient.tests import utils
from mls.apiclient.utils import join_url


class BaseTestCase(unittest.TestCase):
    """Test case base class."""
    HOST = 'demomls.com'
    PATH = '/api'

    @property
    def BASE_URL(self):
        """Return the MLS base url used for the tests."""
        return 'http://{0}'.format(self.HOST)

    @property
    def URL(self):
        """Return the MLS API endpoint url for the tests."""
        return self.BASE_URL + self.PATH

    @property
    def API_BASE(self):
        """Return the base API url for the tests."""
        return join_url(self.BASE_URL, REST_API_URL, REST_API_VERSION)

    def setup_integration_test(self):
        """Setup all URL mocks to run a full integration test."""

        def _register(endpoint, content=None, fixture=None, params=None):
            if fixture:
                content = utils.load_fixture(fixture)
            base_url = utils.get_url(self.API_BASE, endpoint)
            if not params:
                httpretty.register_uri(
                    httpretty.GET,
                    re.compile(base_url),
                    body=content,
                    match_querystring=True,
                )
            else:
                for keys in itertools.permutations(params.keys()):
                    query = urllib.urlencode(
                        [(key, params.get(key)) for key in keys]
                    )
                    httpretty.register_uri(
                        httpretty.GET,
                        re.compile('\?'.join((base_url, query))),
                        body=content,
                        match_querystring=True,
                    )

        base_params = {
            'apikey': 'YOUR_API_KEY',
            'lang': 'en',
        }

        # register all the field endpoints
        _register(
            'field_titles/developments',
            params=base_params,
            fixture='development_fields_en.json',
        )
        _register(
            'field_order/developments',
            params=base_params,
            fixture='development_fields_order.json',
        )
        _register(
            'field_titles/development_groups',
            params=base_params,
            fixture='group_fields_en.json',
        )
        _register(
            'field_order/development_groups',
            params=base_params,
            fixture='group_fields_order.json',
        )
        _register(
            'field_titles/development_phases',
            params=base_params,
            fixture='phase_fields_en.json',
        )
        _register(
            'field_order/development_phases',
            params=base_params,
            fixture='phase_fields_order.json',
        )

        # register the development endpoints
        _register(
            'developments',
            params=base_params,
            fixture='integration/development_list_26-1.json',
        )
        _register(
            'developments',
            params=dict({'limit': 25, 'offset': 25, }, **base_params),
            fixture='integration/development_list_26-2.json',
        )
        _register(
            'developments',
            params=dict({'agency_developments': 'dev-agency'}, **base_params),
            fixture='integration/development_list_15-agency1.json',
        )
        _register(
            'developments',
            params=dict({'agency_developments': 'budget-dev'}, **base_params),
            fixture='integration/development_list_11-agency2.json',
        )



# -*- coding: utf-8 -*-
"""Python client for the RESTful API of the Propertyshelf MLS."""


HTTP_HEADER_PREFIX = 'X-MLS'
PRODUCT_NAME = 'mls.apiclient'
REST_API_URL = 'api/rest'
REST_API_VERSION = 'v1'



# -*- coding: utf-8 -*-
"""Test API class."""

# python imports
import httpretty
import requests

# local imports
from mls.apiclient import api, exceptions
from mls.apiclient.tests import base, utils


class TestAPI(base.BaseTestCase):
    """API test case."""
    PATH = '/api/rest/v1/developments'

    def setUp(self):
        self.api = self._callFUT(self.BASE_URL)

    def _callFUT(self, base_url, api_key=None, lang=None, debug=False):
        return api.API(base_url, api_key=api_key, lang=lang, debug=debug)

    def test_class(self):
        """Validate the class initialization and attributes."""
        api = self._callFUT(
            self.BASE_URL,
            api_key='1234567890abcdef',
            lang='de',
            debug=True,
        )
        self.assertEqual(api.base_url, self.BASE_URL)
        self.assertEqual(api.api_key, '1234567890abcdef')
        self.assertEqual(api.lang, 'de')
        self.assertTrue(api.debug)

    def test_headers(self):
        """Validate the ``headers`` method."""
        headers = self.api.headers()
        self.assertEqual(len(headers.keys()), 2)
        self.assertEqual(headers['Content-Type'], 'application/json')
        self.assertEqual(headers['Accept'], 'application/json')

    @httpretty.httprettified
    def test_handle_http_response_200(self):
        """Validate a HTTP 200 code."""
        content = u'{"some": "content"}'
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            status=200,
        )
        response = requests.get(self.URL)
        result = self.api.handle_response(response, content)
        data = result.get('response')
        self.assertEqual(data, {'some': 'content'})

        result = self.api.handle_response(response, None)
        data = result.get('response')
        self.assertEqual(data, {})

        self.assertRaises(
            ValueError,
            self.api.handle_response, response, u'No JSON',
        )

    @httpretty.httprettified
    def test_handle_api_response_200(self):
        """Validate a API 200 code."""
        content = utils.wrap_content(u'{"some": "content"}', status_code=200)
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        response = requests.get(self.URL)
        result = self.api.handle_response(response, content)
        data = result.get('response')
        self.assertEqual(data, {'some': 'content'})

    @httpretty.httprettified
    def test_handle_http_response_30x(self):
        """Validate a HTTP 30x code."""
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            status=301,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.Redirection,
            self.api.handle_response, response, None,
        )

    @httpretty.httprettified
    def test_handle_api_response_30x(self):
        """Validate a API 30x code."""
        content = utils.wrap_content(u'"Redirect"', status_code=301)
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.Redirection,
            self.api.handle_response, response, content,
        )

    @httpretty.httprettified
    def test_handle_http_response_400(self):
        """Validate a HTTP 400 code."""
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            status=400,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.BadRequest,
            self.api.handle_response, response, None,
        )

    @httpretty.httprettified
    def test_handle_api_response_400(self):
        """Validate a API 400 code."""
        content = utils.wrap_content(u'"Bad request"', status_code=400)
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.BadRequest,
            self.api.handle_response, response, content,
        )

    @httpretty.httprettified
    def test_handle_http_response_401(self):
        """Validate a HTTP 401 code."""
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            status=401,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.UnauthorizedAccess,
            self.api.handle_response, response, None,
        )

    @httpretty.httprettified
    def test_handle_api_response_401(self):
        """Validate a API 401 code."""
        content = utils.wrap_content(u'"Unauthorized access"', status_code=401)
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.UnauthorizedAccess,
            self.api.handle_response, response, content,
        )

    @httpretty.httprettified
    def test_handle_http_response_500(self):
        """Validate a HTTP 500 code."""
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            status=500,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.ServerError,
            self.api.handle_response, response, None,
        )

    @httpretty.httprettified
    def test_handle_api_response_500(self):
        """Validate a API 500 code."""
        content = utils.wrap_content(u'"Server Error"', status_code=500)
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        response = requests.get(self.URL)
        self.assertRaises(
            exceptions.ServerError,
            self.api.handle_response, response, content,
        )

    @httpretty.httprettified
    def test_get(self):
        """Validate the ``get`` method."""
        content = u'{"some": "content"}'
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        result = self.api.get(self.PATH)
        data = result.get('response')
        self.assertEqual(data, {'some': 'content'})

    @httpretty.httprettified
    def test_http_call(self):
        """Validate the ``http_call`` method."""
        content = u'{"some": "content"}'
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        result = self.api.http_call(self.URL, 'GET')
        data = result.get('response')
        self.assertEqual(data, {'some': 'content'})

    @httpretty.httprettified
    def test_request(self):
        """Validate the ``request`` method."""
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body='{}',
            status=200,
        )
        result = self.api.request(self.URL, 'GET')
        data = result.get('response')
        self.assertEqual(data, {})

        content = u'{"some": "content"}'
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        result = self.api.request(self.URL, 'GET')
        data = result.get('response')
        self.assertEqual(data, {'some': 'content'})

    @httpretty.httprettified
    def test_api_debug(self):
        """Validate the API with debug parameter set to True."""
        content = u'{"debug": true}'
        api_debug = self._callFUT(self.BASE_URL, debug=True)
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        result = api_debug.request(self.URL, 'GET')
        data = result.get('response')
        self.assertEqual(data, {'debug': True})

    @httpretty.httprettified
    def test_enveloped_headers(self):
        """Validate the headers returned from the API when they are enveloped
        with the API response.
        """
        content = u'{' + \
            u'"status": 200,' + \
            u'"headers":{"h1": "v1", "h2": "v2"},' + \
            u'"response": {"some": "content"}}'
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            status=200,
        )
        result = self.api.request(self.URL, 'GET')
        headers = result.get('headers')
        self.assertEqual(headers, {'h1': 'v1', 'h2': 'v2'})

    @httpretty.httprettified
    def test_non_enveloped_headers(self):
        """Validate the headers returned from the API when they are not
        enveloped and only included as HTTP headers.
        """
        content = u'{"some": "content"}'
        headers = {
            'X-MLS-h1': 'v1',
            'X-MLS-h2': 'v2',
            'OtherHeader': 'v3',
        }
        httpretty.register_uri(
            httpretty.GET,
            self.URL,
            body=content,
            adding_headers=headers,
            status=200,
        )
        result = self.api.request(self.URL, 'GET')
        response_headers = result.get('headers')
        self.assertEqual(response_headers, {'h1': 'v1', 'h2': 'v2'})



# -*- coding: utf-8 -*-
"""Test custom exception classes."""

# python imports
from collections import namedtuple

# local imports
from mls.apiclient import exceptions
from mls.apiclient.tests import base


class TestExceptions(base.BaseTestCase):

    def setUp(self):
        self.Response = namedtuple('Response', 'status_code reason')

    def test_connection(self):
        error = exceptions.ConnectionError(
            400,
            reason='Bad request',
            url='http://demomls.com/request',
        )
        self.assertEqual(
            str(error),
            'Response status: 400. Reason: Bad request. '
            'URL: http://demomls.com/request',
        )

    def test_redirect(self):
        error = exceptions.Redirection(
            401,
            reason='Redirect => http://example.com',
        )
        self.assertEqual(
            str(error),
            'Response status: 401. Reason: Redirect => http://example.com.',
        )

    def test_not_found(self):
        response = self.Response(status_code='404', reason='Not Found')
        error = exceptions.ResourceNotFound(
            response.status_code,
            response.reason,
        )
        self.assertEqual(
            str(error),
            'Response status: {0}. Reason: {1}.'.format(
                response.status_code, response.reason,
            )
        )

    def test_unauthorized_access(self):
        response = self.Response(status_code='401', reason='Unauthorized')
        error = exceptions.UnauthorizedAccess(
            response.status_code,
            response.reason,
        )
        self.assertEqual(
            str(error),
            'Response status: {0}. Reason: {1}.'.format(
                response.status_code, response.reason,
            )
        )

    def test_missing_param(self):
        error = exceptions.MissingParam('Missing Payment Id')
        self.assertEqual(str(error), 'Missing Payment Id')

    def test_missing_config(self):
        error = exceptions.MissingParam('Missing client_id')
        self.assertEqual(str(error), 'Missing client_id')



# -*- coding: utf-8 -*-
"""Setup for mls.apiclient package."""

import os
from setuptools import setup, find_packages

setup(
    name='mls.apiclient',
    version='1.3',
    description="Python client for the RESTful API of the Propertyshelf MLS.",
    long_description='\n\n'.join([
        open("README.rst").read() + "\n" +
        open(os.path.join("docs", "HISTORY.txt")).read(),
        open(os.path.join("docs", "INSTALL.txt")).read(),
        open(os.path.join("docs", "LICENSE.txt")).read(),
    ]),
    # Get more strings from
    # http://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    keywords='MLS API client',
    author='Propertyshelf, Inc.',
    author_email='development@propertyshelf.com',
    url='https://github.com/propertyshelf/mls.apiclient',
    download_url='http://pypi.python.org/pypi/mls.apiclient',
    license='GPL',
    packages=find_packages('src', exclude=['ez_setup']),
    package_dir={'': 'src'},
    namespace_packages=['mls'],
    include_package_data=True,
    zip_safe=False,
    extras_require=dict(
        test=[
            'httpretty',
            'unittest2',
        ],
    ),
    install_requires=[
        'setuptools',
        'requests',
    ],
    entry_points="""""",
)



# -*- coding: utf-8 -*-

# python imports
import datetime
import json
import logging
import requests

# local imports
from mls.apiclient import HTTP_HEADER_PREFIX, PRODUCT_NAME
from mls.apiclient import exceptions, utils

logger = logging.getLogger(PRODUCT_NAME)


class API(object):
    """API class for the MLS.

    "The class contains all the base methods for establishing and
    authenticating the connections and requests to the MLS database.
    """

    def __init__(self, base_url, api_key=None, lang=None, debug=None):
        """Create API object.

        Usage::

            >>> from mls.apiclient import api
            >>> mls = api.API('http://demomls.com')
        """
        self.base_url = base_url
        self.api_key = api_key
        self.lang = lang
        self.debug = debug

    def request(self, url, method, body=None, params=None):
        """Make HTTP call, formats response and does error handling.
        Uses http_call method in API class.
        """

        url, url_params = utils.split_url_params(url)
        if params:
            url_params = utils.merge_dict(url_params, params)
        if self.api_key:
            url_params = utils.merge_dict(url_params, {'apikey': self.api_key})
        if self.lang:
            url_params = utils.merge_dict(url_params, {'lang': self.lang})
        url = utils.join_url_params(url, url_params)

        try:
            return self.http_call(
                url,
                method,
                data=json.dumps(body),
                headers=self.headers(),
            )
        except exceptions.BadRequest as error:
            # Format Error message for bad request
            return {'error': json.loads(error.content)}
        except requests.ConnectionError:
            raise exceptions.ServerError(503, url=url)

    def http_call(self, url, method, **kwargs):
        """Makes a http call. Logs response information."""

        if self.debug:
            logger.info('Request[{0}]: {1}'.format(method, url))
        start_time = datetime.datetime.now()

        response = requests.request(
            method,
            url,
            **kwargs
        )

        duration = datetime.datetime.now() - start_time
        if self.debug:
            logger.info('Response[{0}]: {1}, Duration: {2}.{3}s.'.format(
                response.status_code,
                response.reason,
                duration.seconds,
                duration.microseconds,
            ))

        return self.handle_response(response, response.content.decode('utf-8'))

    def handle_response(self, response, content):
        """Validate HTTP response."""

        def _validate_status_code(status_code, msg=None, url=None):
            if 200 <= status_code <= 299:
                return True
            elif status_code in (301, 302, 303, 307):
                raise exceptions.Redirection(status_code, msg, url)
            elif status_code == 400:
                raise exceptions.BadRequest(status_code, msg, url)
            elif status_code == 401:
                raise exceptions.UnauthorizedAccess(status_code, msg, url)
            elif status_code == 404:
                raise exceptions.ResourceNotFound(status_code, msg, url)
            elif 500 <= status_code <= 599:
                raise exceptions.ServerError(status_code, msg, url)
            else:
                raise exceptions.ConnectionError(status_code, msg, url)
            return False

        status = response.status_code
        msg = response.reason
        url = response.url
        if _validate_status_code(status, msg, url):
            data = json.loads(content) if content else {}
            status = data.get('status', None)
            if status is None:
                return utils.wrap_data_response(
                    data,
                    headers=utils.extract_headers(
                        response.headers,
                        HTTP_HEADER_PREFIX,
                    )
                )
            msg = data.get('response', '')
            if _validate_status_code(status, msg, url):
                return data

    def headers(self):
        """Default HTTP headers."""

        return {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        }

    def get(self, action, params=None):
        """Make GET request."""

        return self.request(
            utils.join_url(self.base_url, action),
            'GET',
            params=params or {},
        )



# -*- coding: utf-8 -*-
"""Test utility methods."""

# local imports
from mls.apiclient import utils
from mls.apiclient.tests import base


class TestUtils(base.BaseTestCase):
    """Util methods Test Case."""

    def test_extract_headers(self):
        headers = {
            'prefix1-header1': 'value1',
            'prefix1-header2': 'value2',
            'prefix2-header3': 'value3',
            'prefix1-header4': 'value4',
        }
        result = utils.extract_headers(headers, 'prefix1')
        self.assertEqual(result, {
            'header1': 'value1',
            'header2': 'value2',
            'header4': 'value4',
        })

        result = utils.extract_headers(headers, 'prefix2')
        self.assertEqual(result, {'header3': 'value3'})

        result = utils.extract_headers(headers, 'prefix3')
        self.assertEqual(result, {})



# -*- coding: utf-8 -*-
"""MLS API Client implementation.

mls.apiclient is a Python client for the RESTful API of the Propertyshelf MLS.
"""
from copy import deepcopy
from urlparse import urljoin
import datetime
import logging
import requests

from mls.apiclient.exceptions import (
    ImproperlyConfigured,
    MLSError,
    ObjectNotFound,
)

API_URL = 'api'

#: Disable noisy messages from requests module.
requests_log = logging.getLogger('requests')
requests_log.setLevel(logging.WARNING)

logger = logging.getLogger('mls.apiclient.client')


class ResourceBase(object):
    """Base resource class."""

    path = None
    path_search = 'search'
    path_detail = 'detail'
    path_categories = None

    def __init__(self, base_url, api_key='', path=None, debug=False):
        self._base_url = base_url
        self._api_key = api_key
        self._debug = debug
        if path:
            self.path = path
        self._url = '/'.join([self._base_url, API_URL, self.path])
        self._format = 'json'

    def all(self):
        """Returns all objects of this Resource."""
        return self._query()

    def get(self, key, lang=None):
        """Returns one object of this Resource.

        You have to give one keyword argument to find the object.
        """
        params = {}
        params['search'] = '/'.join([self.path_detail, key])
        if lang is not None:
            params['lang'] = lang
        result = self._query(params, batching=False)
        if result is None:
            raise ObjectNotFound('Item not found.')
        return result

    def category(self, key, lang=None):
        """Return values for a categorie."""
        if self.path_categories is None:
            return
        params = {}
        params['search'] = '/'.join([self.path_categories, key])
        if lang is not None:
            params['lang'] = lang
        result = self._query(params, batching=False)
        if result is None:
            raise ObjectNotFound('Item not found.')
        return [tuple(item) for item in result]

    def search(self, params):
        """Returns a list of objects.

        You can search for objects by giving one or more keyword arguments.
        Use limit and offset to limit the results.
        """
        if len(params) == 0:
            raise MLSError('You have to give at least one search argument.')
        params['search'] = self.path_search
        results, batching = self._query(params)
        return results, batching

    def _get_response(self, url, params, timeout=2.5):
        """Get the response from the MLS.

        :param url: [required] Request URL.
        :type url: string
        :param params: [required] Request params.
        :type params: dict
        :param timeout: Request timeout.
        :type timeout: float
        :returns: response
        :rtype: requests.response object
        """
        if self._debug:
            logger.info('Request: {0}'.format(url))
            start_time = datetime.datetime.now()

        try:
            r = requests.get(url, params=params, timeout=timeout)
        except requests.exceptions.ConnectionError, e:
            raise MLSError(
                'Connection to the MLS at {0} failed.'.format(e.request.url)
            )
        except requests.exceptions.MissingSchema:
            raise MLSError(
                'No or wrong MLS URL provided.'
            )
        except requests.exceptions.Timeout, e:
            raise MLSError(
                'Connection to the MLS at {0} timed out.'.format(e.request.url)
            )

        if self._debug:
            duration = datetime.datetime.now() - start_time
            logger.info('Response[{0}]: {1}, Duration: {2}.{3}s.'.format(
                r.status_code,
                r.reason,
                duration.seconds,
                duration.microseconds,
            ))
        return r

    def _query(self, _params, batching=True):
        """Generates the URL and sends the HTTP request to the MLS."""
        url = self._url
        params = deepcopy(_params)
        search = params.pop('search', '')
        if search:
            url = urljoin(url + '/', search)
        params['apikey'] = self._api_key
        params['format'] = self._format
        # encoded_args = urllib.urlencode(params)
        # url = url + '?' + encoded_args

        r = self._get_response(url, params)

        try:
            response = r.json()
        except ValueError:
            raise MLSError(
                'The data returned from the server could not be read.'
            )

        if response.get('status', None) != 'ok':
            raise ImproperlyConfigured('Wrong request ({0}).'.format(r.url))

        results = response.get('result', None)
        if not batching:
            return results

        batching = response.get('batching', None)
        batch = None
        if batching and batching.get('active', False):
            batch = {
                'results': batching.get('results_total', 0),
                'items': batching.get('results_page', 0),
                'next': batching.get('next', None),
                'prev': batching.get('prev', None)
            }
        return results, batch


class ListingResource(ResourceBase):
    """Listings Resource."""
    path = 'listings'
    path_search = 'search'
    path_detail = 'listing'
    path_categories = 'categories'



# -*- coding: utf-8 -*-

# python imports
import urlparse
import re

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode


def join_url(url, *paths):
    """Joins individual URL strings together, and returns a single string.

    Usage::

        >>> join_url('example.com', 'index.html')
        'example.com/index.html'
    """
    for path in paths:
        url = re.sub(r'/?$', re.sub(r'^/?', '/', path), url)
    return url


def join_url_params(url, params):
    """Constructs percent-encoded query string from given params dictionary.

    The result is appended to the given url.

    Usage::

        >>> join_url_params('example.com/index', {"count": 5, "type": "rl"})
        example.com/index?count=5&type=rl
    """
    return url + '?' + urlencode(params)


def merge_dict(data, *override):
    """Merges any number of dictionaries together.

    The result is a single dictionary.

    Usage::

        >>> merge_dict ({"foo": "bar"}, {1: 2}, {"count": 5})
        {1: 2, 'foo': 'bar', 'count': 5}
    """
    result = {}
    for current_dict in (data,) + override:
        result.update(current_dict)
    return result


def get_link(links, name):
    """Returns a specific link from a list of links with the given format:

    Usage::

        >>> links = [
        ...     {
        ...         "href" : "http://demomls.com",
        ...         "rel" : "self",
        ...     },
        ...     {
        ...         "href" : "http://demomls.com/next",
        ...         "rel" : "next",
        ...     },
        ... ]
        >>> get_link(links, 'self')
        http://demomls.com
        >>> get_link(links, 'next')
        http://demomls.com/next
    """
    if links is None:
        return None
    for link in links:
        if link.get('rel') == name:
            return link.get('href')


def split_url_params(url):
    """Split a url with parameters to return a tuple of the base url and a
    dictionary of the parameters.

    Usage::

        >>> split_url_params('http://demomls.com?param1=value1&param2=value2')
        ('http://demomls.com', {'param1': 'value1', 'param2': 'value2'})
    """
    params = {}
    parsed = url.split('?')
    base_url = parsed[0]
    if len(parsed) > 0:
        result = urlparse.urlparse(url)
        params = dict(urlparse.parse_qsl(result.query))
    return (base_url, params)


def extract_headers(headers, prefix):
    """Extract HTTP headers with the specific prefix from the HTTP response
    and return them in a dict.
    """
    prefix = prefix.upper() + '-'
    return dict((key[len(prefix):], val)
                for key, val in headers.items()
                if key.upper().startswith(prefix))


def wrap_data_response(data, status=200, headers={}):
    """Wrap a data dictionary with the appropriate response format if the
    response uses enveloped headers with optional parameters to include a
    specific status code and/or headers.
    """
    return {
        'status': status,
        'headers': headers,
        'response': data,
    }
